from antlr4 import *
from schema_gen.antlr import GraphQLLexer, GraphQLListener, GraphQLParser
from schema_gen.codegen import CodegenTool, Class, String, ClassInstance, IfElse, If, Method, Expr, Variable
import re
from math import floor
from datetime import datetime
from schema_gen.utils import strip_string_quotes, camel_case_to_snake_case, process_input_value_definition
from schema_gen.errors import ParsingError

GraphQLParser = GraphQLParser.GraphQLParser

graphene = 'graphene'
built_in_scalars = [
    'Int',
    'Float',
    'String',
    'Boolean',
    'ID',
    'Date',
    'Datetime',
    'Time'
    'Decimal',
    'JSONString',
    'Base64',
]


class SDLParser(GraphQLListener.GraphQLListener):
    def __init__(self, input_file, output_file=None):
        if output_file is None:
            output_file = input_file.split(sep='.')[0] + '_' + str(floor(datetime.now().timestamp())) + '.py'

        is_valid_file_name = re.match("\w+.py$", output_file)
        if is_valid_file_name is None:
            raise Exception('File is not a python file')
        self.output_file = output_file
        self.input_file = input_file
        self.codegen = CodegenTool(output_file=self.output_file)
        super().__init__()

    def enterTypeDefinition(self, ctx: GraphQLParser.TypeDefinitionContext):

        for child in ctx.children:
            # type definition is for an Object Type Definition
            if isinstance(child, GraphQLParser.ObjectTypeDefinitionContext) or isinstance(child,
                                                                                          GraphQLParser.InterfaceTypeDefinitionContext):
                # print(child.name().getText())
                # if ":" in child.name().getText():
                #     child.exitRule(self)
                is_object_type = isinstance(child, GraphQLParser.ObjectTypeDefinitionContext)
                is_interface = isinstance(child, GraphQLParser.InterfaceTypeDefinitionContext)

                type_class = Class(name=child.name().getText(), add_init_method=False)
                if is_object_type:
                    type_class.base_class = "ObjectType"
                elif is_interface:
                    type_class.base_class = "Interface"

                is_mutation = False
                if type_class.name == 'Mutation':
                    is_mutation = True
                    is_object_type = False

                meta_class = Class(name='meta')

                # create map for methods to be resolved
                methods_to_be_resolved = {}

                # get type description
                desc = child.description()
                if desc:
                    meta_class.add_class_variable('description', String(strip_string_quotes(desc.getText())))

                # get implemented interfaces
                if is_object_type or is_mutation:
                    if child.implementsInterfaces() is not None:
                        interfaces = child.implementsInterfaces().getText().split(sep='implements')
                        interfaces = interfaces[1].split(sep='&')
                        interface_string = ''
                        for i in interfaces:
                            interface_string = interface_string + i + ','
                        meta_class.add_class_variable('interfaces', f"({interface_string})")

                # get fields of the ObjectType or Interface

                if child.fieldsDefinition():
                    fields = child.fieldsDefinition().fields
                    if not is_mutation:
                        for field in fields:
                            # get field name and type
                            field_name = camel_case_to_snake_case(field.name().getText())
                            field_type = field.type_().getText()
                            field_required = False

                            # get field description
                            field_desc = field.description()
                            if field_desc is not None:
                                field_desc = String(strip_string_quotes(field_desc.getText()))
                            else:
                                field_desc = ''

                            if is_interface:
                                if field_type.lower() == type_class.name.lower():
                                    field_type = 'lambda: ' + field_type

                            # if field is a required field
                            if field_type[len(field_type) - 1] == '!':
                                field_required = True
                                field_code = ClassInstance('Field', field_type[:-1], required=True)
                            else:
                                field_code = ClassInstance('Field', field_type)

                            # if field is a list type
                            if field.type_().listType() is not None:
                                list_type_named_type = field.type_().listType().type_().getText()

                                if is_interface:
                                    if list_type_named_type.lower() == type_class.name.lower():
                                        list_type_named_type = 'lambda: ' + list_type_named_type

                                if list_type_named_type[len(list_type_named_type) - 1] == '!':
                                    field_code = ClassInstance('List',
                                                               str(ClassInstance('NonNull', list_type_named_type[:-1])),
                                                               required=field_required)
                                else:
                                    field_code = ClassInstance('List', list_type_named_type, required=field_required)

                            # get field arguments
                            if is_object_type:
                                args = field.argumentsDefinition()
                                args_string = []
                                if args is not None:
                                    args = args.args
                                    for arg in args:
                                        # add info to method_to_be_resolved map
                                        if field_name not in methods_to_be_resolved:
                                            methods_to_be_resolved[field_name] = [arg.name().getText()]
                                        else:
                                            methods_to_be_resolved[field_name].append(arg.name().getText())
                                        processed_arg = process_input_value_definition(arg)
                                        args_string.append(
                                            f"{String(processed_arg['name'])}: {str(processed_arg['arg_impl'])}")

                                    field_code.add_kwarg('args', "{" + ', '.join(args_string) + "}")

                            if field_desc != '':
                                field_code.add_kwarg(key='description', value=field_desc)
                            type_class.class_variables[field_name] = str(field_code)
                    else:
                        for field in fields:
                            # get field name and type
                            field_name = camel_case_to_snake_case(field.name().getText())
                            field_type = field.type_().getText()
                            field_required = False

                            field_class = Class(field.name().getText(), add_init_method=False, base_class='Mutation')
                            argument_class = Class(name='arguments')

                            # get field description
                            field_desc = field.description()
                            if field_desc is not None:
                                field_desc = String(strip_string_quotes(field_desc.getText()))
                            else:
                                field_desc = ''

                            # if field is a required field
                            if field_type[len(field_type) - 1] == '!':
                                field_required = True
                                field_code = ClassInstance('Field', field_type[:-1], required=True)
                            else:
                                field_code = ClassInstance('Field', field_type)

                            # if field is a list type
                            if field.type_().listType() is not None:
                                list_type_named_type = field.type_().listType().type_().getText()

                                if list_type_named_type[len(list_type_named_type) - 1] == '!':
                                    field_code = ClassInstance('List',
                                                               str(ClassInstance('NonNull', list_type_named_type[:-1])),
                                                               required=field_required)
                                else:
                                    field_code = ClassInstance('List', list_type_named_type, required=field_required)

                            # get field arguments
                            args = field.argumentsDefinition()
                            arg_list = []
                            if args is not None:
                                args = args.args
                                for arg in args:
                                    processed_arg = process_input_value_definition(arg)
                                    argument_class.add_class_variable(processed_arg['name'],
                                                                      str(processed_arg['arg_impl']))
                                    arg_list.append(processed_arg['name'])

                            field_class.add_sub_class(argument_class)
                            field_class.add_method(
                                method=Method(
                                    name='mutate',
                                    arguments=['root', 'info'] + arg_list
                                )
                            )

                            if field_desc != '':
                                field_code.add_kwarg(key='description', value=field_desc)

                            # write mutation classes for the mutation's fields
                            self.codegen.write_class(field_class)
                            type_class.class_variables[field_name] = str(field_code)

                # add resolver methods
                if not is_mutation:
                    for method in methods_to_be_resolved:
                        type_class.add_method(method_name='resolve_' + method,
                                              arguments_names=['info'] + methods_to_be_resolved[method])

                if type_class.name == 'Query':
                    for var in type_class.class_variables:
                        if var not in methods_to_be_resolved:
                            type_class.add_method(method_name='resolve_' + var, arguments_names=['info'])

                if len(meta_class.class_variables) != 0:
                    type_class.add_sub_class(meta_class)

                self.codegen.write_class(type_class)

            # type definition is for an EnumTypeDefinition
            elif isinstance(child, GraphQLParser.EnumTypeDefinitionContext):
                enum_class = Class(name=child.name().getText(), base_class="Enum", add_init_method=False)
                meta_class = Class(name='meta')

                # get enum description
                desc = child.description()
                if desc:
                    meta_class.add_class_variable('description', String(strip_string_quotes(desc.getText())))

                # get fields of the Enum
                fields = child.enumValuesDefinition().fields
                fields_and_desc = {}
                for field in fields:
                    # get field name and type
                    enum_value = field.enumValue().getText()

                    # get enum description
                    field_desc = field.description()
                    if field_desc is not None:
                        field_desc = String(strip_string_quotes(field_desc.getText()))
                    else:
                        field_desc = ''

                    if field_desc != '':
                        # do something
                        fields_and_desc[enum_value] = field_desc

                    # add enums as class variables to main class
                    enum_class.add_class_variable(enum_value, String(enum_value))

                # add enums description
                method = Method(
                    name='description',
                    decorators=['@property'],
                    arguments=[]
                )

                if_else = IfElse(
                    indent_level=method.get_indent_level() + 1,
                    else_action=[Expr("pass")],
                )

                for i in fields_and_desc:
                    if_else.add_elif(If(
                        expr=Expr(f"self == {enum_class.name}.{i}"),
                        action=[Expr(f"return {fields_and_desc[i]}")]
                    ))

                method.set_body([if_else])

                enum_class.add_method(method=method)

                if len(meta_class.class_variables) != 0:
                    enum_class.add_sub_class(meta_class)

                self.codegen.write_class(enum_class)

            # type definition is for an EnumTypeDefinition
            elif isinstance(child, GraphQLParser.ScalarTypeDefinitionContext):
                if child.name().getText().capitalize() in built_in_scalars:
                    continue
                scalar_class = Class(name=child.name().getText(), base_class="Scalar", add_init_method=False)
                desc = child.description()

                if desc is not None:
                    scalar_class.description = strip_string_quotes(desc.getText())

                serialize_method = Method(
                    name='serialize',
                    arguments=['val'],
                    decorators=['@staticmethod'],
                    body=[Expr('# write method body'), Expr('pass')],
                    is_static=True
                )
                parse_literal_method = Method(
                    name='parse_literal',
                    arguments=['node'],
                    decorators=['@staticmethod'],
                    body=[Expr('# write method body'), Expr('pass')],
                    is_static=True
                )

                parse_value_method = Method(
                    name='parse_value',
                    arguments=['value'],
                    decorators=['@staticmethod'],
                    body=[Expr('# write method body'), Expr('pass')],
                    is_static=True
                )

                scalar_class.add_method(method=serialize_method)
                scalar_class.add_method(method=parse_literal_method)
                scalar_class.add_method(method=parse_value_method)

                self.codegen.write_class(scalar_class)

            elif isinstance(child, GraphQLParser.UnionTypeDefinitionContext):
                union_class = Class(name=child.name().getText(), base_class='Union')
                meta_class = Class(name='Meta')

                unions = child.unionMemberTypes().getText()

                if unions[0] == '=':
                    unions = unions[1:]

                unions = ', '.join(unions.split(sep='|'))

                meta_class.add_class_variable(variable_name='types', variable_value=f"({unions})")

                desc = child.description()

                if desc is not None:
                    meta_class.add_class_variable(variable_name='description',
                                                  variable_value=String(strip_string_quotes(desc.getText())))

                union_class.add_sub_class(meta_class)
                self.codegen.write_class(union_class)
                print(unions)

            elif isinstance(child, GraphQLParser.InputObjectTypeDefinitionContext):
                type_class = Class(name=child.name().getText(), base_class="InputObjectType", add_init_method=False)
                meta_class = Class(name='meta')

                # get type description
                desc = child.description()
                if desc:
                    meta_class.add_class_variable('description', String(strip_string_quotes(desc.getText())))

                # get fields
                if child.inputFieldsDefinition():
                    fields = child.inputFieldsDefinition().fields
                    for field in fields:
                        processed_ivd = process_input_value_definition(field)
                        type_class.add_class_variable(processed_ivd['name'], str(processed_ivd['arg_impl']))

                if len(meta_class.class_variables) != 0:
                    type_class.add_sub_class(meta_class)

                self.codegen.write_class(type_class)
            else:
                print(type(child))

    def enterSchemaDefinition(self, ctx: GraphQLParser.SchemaDefinitionContext):

        schema_obj = ClassInstance('Schema')

        fields = ctx.fields
        for field in fields:
            schema_obj.add_kwarg(strip_string_quotes(field.operationType().getText()),
                                 strip_string_quotes(field.namedType().getText()))

        var = Variable(
            name='schema',
            value=schema_obj
        )
        self.codegen.write_variable(var)

    def __call__(self):
        try:
            self.codegen.import_package(package=graphene, mode=2, object='*')
            input_stream = FileStream(self.input_file)
            lexer = GraphQLLexer.GraphQLLexer(input_stream)
            stream = CommonTokenStream(lexer)
            parser = GraphQLParser(stream)
            tree = parser.document()
            walker = ParseTreeWalker()
            walker.walk(self, tree)
        except Exception as err:
            raise ParsingError(str(err))



# TODO: Fix issue with parser reading type in fields
# TODO: Write documentation
# TODO: Publish app



import datetime
from math import floor
import autopep8
import re
from typing import List
from schema_gen.utils import stringify, triple_stringify
from schema_gen.errors import CodegenError


class Block:
    """
    Block is an 'interface' that represents a block of code. It is mostly used for indenting
    """

    def __init__(self, indent_level):
        self.indent_level = indent_level

    def set_indent_level(self, indent_level: int):
        self.indent_level = indent_level

    def get_indent_level(self) -> int:
        return self.indent_level

    def increment_indent_level(self):
        self.indent_level = self.indent_level + 1

    def __str__(self):
        pass


# The String type is used to differentiate between regular strings and strings that represent code
class String:
    def __init__(self, string):
        self.value = string

    def __str__(self):
        return stringify(self.value)


class Expr(Block):
    """
    This simple class represents a python expression e.g: a < b
    """

    def __init__(self, expression: str, indent_level=0):
        self.expression = expression
        super().__init__(indent_level)

    def __str__(self):
        tabs = '\t' * self.indent_level
        return tabs + self.expression


class Function(Block):
    def __init__(self, name: str, arguments: list[str], body: List[Block] = None, indent_level=0,
                 decorators: List[str] = None):
        """
        This is used to initialise a Function object
        Args:
            name: name of the function
            arguments: arguments that the function take
            indent_level: indent level of the function
            decorators: decorators applied to the function. Each decorator should be in the form '@<decorator_name>'
        """
        self.name = name
        self.arguments = arguments
        if decorators is None:
            decorators = []
        self.decorators = decorators
        self.body = body
        super().__init__(indent_level)

    def __str__(self):
        args = ""
        if len(self.arguments) != 0:
            args = ', '.join(self.arguments)
        tabs = '\t' * self.get_indent_level()
        decorators = ''
        for i in self.decorators:
            decorators = decorators + f"\n{tabs}{i}"

        if self.body is None:
            return f"{decorators}\n{tabs}def {self.name}({args}):\n{tabs}\t# write method body\n{tabs}\tpass"
        else:
            body_lines = ''
            for block in self.body:
                block.set_indent_level(self.get_indent_level() + 1)
                if isinstance(block, Expr):
                    body_lines = body_lines + '\n' + str(block)
                    continue
                body_lines = body_lines + str(block)

            return f"{decorators}\n{tabs}def {self.name}({args}):{body_lines}"

    def add_decorator(self, decorator: str):
        if decorator[0] != '@':
            raise CodegenError('Invalid decorator value. Decorator should start with @')
        else:
            self.decorators.append(decorator)

    def set_body(self, body: List[Block]):
        self.body = body


# TODO: make sure everything works well after sha
class Method(Function):
    def __init__(self, name: str, arguments: list[str], indent_level=1, decorators: List[str] = None,
                 body: List[Block] = None, is_static: bool = False):
        if not is_static:
            arguments = ['self'] + arguments
        super().__init__(name, arguments, indent_level=indent_level, decorators=decorators, body=body)


class Class(Block):
    def __init__(self, name: str, base_class: str = None, add_init_method: bool = False, indent_level=0,
                 description: str = None):
        self.name = name[0].upper()+name[1:]
        self.base_class = base_class
        self.add_init_method = add_init_method
        self.methods = []
        self.class_variables = {}
        self.subclasses = []
        self.description = description
        super().__init__(indent_level)

    def __str__(self):
        class_def = ""
        init = ""
        methods = ""
        class_vars = ""
        subclasses = ""

        description = ''
        if self.description != '':
            description = '\n\t' + self.description

        # get tabs for indenting
        tabs = '\t' * self.get_indent_level()
        if self.base_class is None:
            class_def = f"\n\n{tabs}class {self.name}:{description}"
        else:
            class_def = f"\n\n{tabs}class {self.name}({self.base_class}):{description}"

        if self.add_init_method is True:
            init = f"\n\t{tabs}def __init__(self):\n\t\t{tabs}# initialise class here\n\t\tpass"

        for method in self.methods:
            methods = methods + f"\n{str(method)}"

        for var in self.class_variables:
            class_vars = class_vars + "\n\t" + tabs + var + " = " + self.class_variables[var]

        for subclass in self.subclasses:
            subclasses = "\t" + str(subclass)

        if subclasses or class_vars or init or methods:
            return f"{class_def}{subclasses}{class_vars}{init}{methods}\n"
        else:
            pass_expr = Expr('pass', indent_level=1)
            return f"{class_def}\n{str(pass_expr)}\n"

    def add_method(self, method_name: str = None, arguments_names: list[str] = None, decorators: List[str] = None,
                   method: Method = None):
        if method is not None:
            method.indent_level = self.get_indent_level() + 1
            self.methods.append(method)
        else:
            if method_name is None or arguments_names is None:
                raise CodegenError('method name or argument name invalid')
            self.methods.append(Method(name=method_name, arguments=arguments_names, decorators=decorators,
                                       indent_level=self.get_indent_level() + 1))

    def add_class_variable(self, variable_name, variable_value):
        value = str(variable_value)
        self.class_variables[variable_name] = value

    def add_sub_class(self, _class):
        # increase the subclass indent by 1 greater than the parent class' indent
        _class.set_indent_level(self.get_indent_level() + 1)
        self.subclasses.append(_class)

    @property
    def description(self):
        return self._description

    @description.setter
    def description(self, description: str):
        if description is None:
            self._description = ''
        else:
            self._description = triple_stringify(description)


class Variable:
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def __str__(self):
        if isinstance(self.value, String) or isinstance(self.value, ClassInstance):
            value = str(self.value)
        else:
            value = self.value
        return "\n" + self.name + " = " + value

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self._value = value


class ClassInstance:
    def __init__(self, class_name, *args, **kwargs):
        """
        Create a ClassInstance object
        Args:
            class_name: the name of the class
            *args: the args passed to the class' init method
            **kwargs: the kwargs passed to the class' init method
        """
        self.class_name = class_name
        self.args = args
        self.kwargs = kwargs

    def __str__(self):
        args = ', '
        kwargs = []

        args = ', '.join(self.args)

        for kwarg in self.kwargs:
            val = self.kwargs[kwarg]
            if isinstance(val, String):
                val = str(val)
            kwargs.append(f"{kwarg}={val}")

        kwargs = ', '.join(kwargs)
        if kwargs != '':
            if args != '':
                kwargs = ', ' + kwargs

        return f"{self.class_name}({args}{kwargs})"

    def add_kwarg(self, key, value):
        self.kwargs[key] = value


class If(Block):
    def __init__(self, expr: Expr, action: List[Expr], if_type: str = 'if', indent_level=0):
        self.expr = expr
        self.action = action
        self.type = if_type
        super().__init__(indent_level)

    def __str__(self):
        actions = ''
        tabs = self.get_indent_level() * '\t'
        for i, action in enumerate(self.action):
            if i == len(self.action) - 1:
                actions = actions + tabs + '\t' + str(action)
            else:
                actions = actions + tabs + '\t' + str(action) + '\n'
        return f"\n{tabs}{self.type} {str(self.expr)}:\n{actions}"


class IfElse(Block):
    def __init__(self, else_action: List[Expr], indent_level=0, elifs: List[If] = None, if_: If = None):
        if elifs is None:
            elifs = []
        self.if_ = if_
        self.else_action = else_action
        self.elifs = elifs
        super().__init__(indent_level)

    def __str__(self):
        if self.if_ is None:
            raise CodegenError('No If object attached to IfElse object')
        tabs = self.get_indent_level() * '\t'
        if_action = ''
        for expr in self.if_.action:
            if_action = if_action + tabs + '\t' + str(expr) + '\n'
        if not if_action:
            if_action = 'pass'

        else_action = ''
        for expr in self.else_action:
            else_action = tabs + '\t' + str(expr) + '\n'
        if not else_action:
            else_action = 'pass'

        elifs = '\n'

        for elif_ in self.elifs:
            elifs = elifs + str(elif_) + '\n'

        if elifs == '\n':
            elifs = ''

        return f"\n{tabs}if {str(self.if_.expr)}:\n{if_action.rstrip()}{elifs}{tabs}else:\n{else_action}"

    def add_elif(self, elif_: If):
        """
        This adds an elif statement to the If Else block. You can add the elif statement manually during
        initialisation of the IfElse object but you would have to set indent levels correctly yourself but this
        method does that for you.

        Args:
            elif_: the elif statement to add

        """
        if self.if_ is None:
            elif_.type = 'if'
            self.if_ = elif_
        else:
            elif_.type = 'elif'
            elif_.set_indent_level(self.indent_level)
            self.elifs.append(elif_)


class CodegenTool:
    """
    This is a little internal tool for programmatic writing of python code
    """

    def __init__(self, output_file: str = None, override: bool = True):
        # create a new codegen tool
        mode = 'a+'
        if override is True:
            mode = 'w'
        if output_file is None:
            output_file = 'codegen_output' + str(floor(datetime.datetime.now().timestamp())) + ".py"
            override = True

        is_valid_file_name = re.match("\w+.py$", output_file)

        if is_valid_file_name is None:
            raise Exception('File is not a python file')
        with open(output_file, mode) as f:
            self.output_file = output_file
            if override is True:
                f.write('# This file was generated by CodegenTool')
            self.format_file()

    def import_package(self, mode=1, **kwargs):
        """
        Imports a package in the codegen output file

        Args:
            mode: import style
                1 - import package
                2 - from package import object
                3 - from package import object as alias
                4 - import package as alias
            **kwargs: package, object. keyword arguments

        """
        if mode == 1:
            statement = f"import {kwargs['package']}"
        elif mode == 2:
            statement = f"from {kwargs['package']} import {kwargs['object']}"
        elif mode == 3:
            statement = f"from {kwargs['package']} import {kwargs['object']} as {kwargs['alias']}"
        elif mode == 4:
            statement = f"import {kwargs['package']} as {kwargs['alias']}"
        else:
            raise Exception("Unrecognised import mode")

        with open(self.output_file, 'a+') as f:
            f.write(f"\n{statement}")
            self.format_file()

    def write_class(self, _class: Class):
        with open(self.output_file, 'a+') as f:
            f.write(str(_class))
            self.format_file()

    def write_variable(self, variable: Variable):
        with open(self.output_file, 'a+') as f:
            f.write(str(variable))
            self.format_file()

    def write_if_else(self, if_else: IfElse):
        with open(self.output_file, 'a+') as f:
            f.write(str(if_else))
            self.format_file()

    def format_file(self):
        autopep8.fix_file(self.output_file)


if __name__ == '__main__':
    # c = Class(name='NewException', base_class='Exception', add_init_method=True)
    # c.add_method(method_name='get_message', arguments_names=[])
    # c.add_method(method_name='another_method', arguments_names=['message'])
    # c.add_class_variable('hey', 2)
    # codegen.import_package(mode=2, package='datetime', object='datetime')
    # codegen.write_class(c)
    # a = ClassInstance('name', String('arg1'), String('arg2'), kwarg1='kwarg10', required='True')
    codegen = CodegenTool('test.py')
    # ie = IfElse(
    #     if_=If(
    #         expr=Expr(f"{String('a')} != {String('b')}"),
    #         action=[Expr("print('true')"), Expr("print('true')"), Expr("print('true')")]
    #     ),
    #     else_action=[Expr("print('else')")],
    # )
    #
    # ie.add_elif(
    #     If(
    #         expr=Expr(f"{String('a')} == {String('b')}"),
    #         action=[Expr("print('false')"), Expr("print('false')"), Expr("print('false')")],
    #         if_type='elif'
    #     )
    # )

    cls = Class(name='randomshii', description='hEY BABE')

    codegen.write_if_else(cls)



class ParsingError(Exception):
    """
    This exception is raised whenever there's an error in parsing the graphql sdl provided to the parser
    """

    def __init__(self, message):
        self.message = message
        super().__init__(message)


class CodegenError(Exception):
    """
    This exception is raised when an error occurs or an invalid value is encountered in the Codegen
    """

    def __init__(self, message):
        self.message = message
        super().__init__(message)



import argparse
import schema_gen

parser = argparse.ArgumentParser(
    prog='schemagen',
    description='Parse a GraphQL Schema file',
)

parser.add_argument(
    'command',
    metavar='command',
    type=str,
    help='command',
    action='store'
)

parser.add_argument(
    'payload',
    metavar='payload',
    type=str,
    help='payload',
    action='store'
)

parser.add_argument(
    '-o',
    '--output',
    metavar='output_file',
    type=str,
    help='Output file for generated code',
    action='store'
)

args = parser.parse_args()
if args.command == 'parse':
    sdl_parser = schema_gen.SDLParser(
        input_file=args.payload,
        output_file=args.output
    )
    sdl_parser()

else:
    print('invalid command!')




def strip_string_quotes(string: str) -> str:
    if string[0] == '"' and string[len(string) - 1] == '"':
        res = string[1:]
        res = res[:-1]
    else:
        return string
    return res


def stringify(string) -> str:
    return "'" + str(string) + "'"


def triple_stringify(string) -> str:
    return '"""' + str(string) + '"""'


def snake_case_to_camel_case(string) -> str:
    parts = string.split(sep='_')
    for i, part in enumerate(parts):
        if i != 0:
            parts[i] = part.capitalize()
    return ''.join(parts)


def camel_case_to_snake_case(string) -> str:
    parts = list(string)
    for i, part in enumerate(parts):
        if part.isupper():
            if i == 0:
                parts[i] = part.lower()
                continue
            parts[i] = '_' + part.lower()
    return ''.join(parts)


def is_a_builtin(word) -> bool:
    """
    Returns True if a word is a builtin python object
    Args:
        word: word to check

    Returns: True if word is builtin python object

    """
    import builtins
    return str(word) in dir(builtins)


def process_input_value_definition(ivd) -> dict:
    """
    This function process an inputValueDefintionContext and returns the needed results
    Args:
        ivd: the inputValueDefinitionContext

    Returns:
        a dict containing the results

    """

    from schema_gen import codegen
    ClassInstance = codegen.ClassInstance

    arg_type = ivd.type_().getText()
    arg_required = False
    if arg_type[len(arg_type) - 1] == '!':
        arg_required = True
        arg_type = arg_type[:-1]

    # this represents the graphene implementation of this IVD (InputValueDefinition)
    arg_impl = ClassInstance(arg_type, required=arg_required)

    # if any of the argument types is a list type:
    if ivd.type_().listType() is not None:
        list_type_named_type = ivd.type_().listType().type_().getText()
        if list_type_named_type[len(list_type_named_type) - 1] == '!':
            arg_impl = ClassInstance('List', str(
                ClassInstance('NonNull', list_type_named_type[:-1])), required=arg_required)
        else:
            arg_impl = ClassInstance('List', list_type_named_type, required=arg_required)

    # get argument description
    if ivd.description() is not None:
        arg_desc = ivd.description().getText()
        arg_impl.add_kwarg('description', arg_desc)

    # get argument default value
    if ivd.defaultValue() is not None:
        arg_impl.add_kwarg('default_value', ivd.defaultValue().value().getText())

    return {
        'name': ivd.name().getText(),
        'arg_impl': arg_impl
    }


if __name__ == '__main__':
    print(camel_case_to_snake_case('CamelCaseToBeFuckingTransformed'))
